#!/usr/local/bin/tclsh8.3

# DocBook Speller
#
# Chern Lee
#    chern@FreeBSD.org

if {[string match "*/*" $argv0]} {
    regexp {(.*)/(.*?)} $argv0 m d
    set d "$d/dictionary.tcl"
} else {
    set d "dictionary.tcl"
}

if {![file readable $d]} {
    puts stderr "Cannot find dictionary.tcl"
    return
} else {
    source "$d"
}

set dictionary [concat $terms $names $locations $months $companies $other \
		       $commands]

set restricted {

}

proc secs { } {
    return [clock seconds]
}

proc speller { line } {
    global ifd

    puts $ifd "$line"
    flush $ifd

    set res ""
    while {$res == ""} {
	set res [gets $ifd]
    }
    switch -- [string index $res 0] {
	"*" { # correct
	    return 1
	}
	"+" { # semi-correct
	    return 1
	}
	"?" { # semi-correct
	    return 1
	}
	"&" { # incorrect
	    return 0
	}
	"\#" { # really wrong
	    return 0
	}
	default {
	    puts "!! $res $line"
	    return 0
	}
    }
}

proc numbertext { text } {
    set ntext [split $text \n]
    set i 1
    foreach line $ntext {
	append ftext "[format "%5s" $i] [string trim $line]\n"
	incr i
    }
    return $ftext
}

proc removetags { text } {
    set reg {
	{<command>.*?</command>}
	{<literallayout>.*?</literallayout>}
	{<literal>.*?</literal>}
	{<programlisting>.*?</programlisting>}
	{<programlisting.*?>.*?</programlisting>}
	{<literallayout .*?>.*?</literallayout>}
	{<screen>.*?</screen>}
	{<userinput>.*?</userinput>}
	{<makevar>.*?</makevar>}
	{<filename>.*?</filename>}
	{<username>.*?</username>}
	{<application>.*?</application>}
	{<indexterm>.*?</indexterm>}
	{<email>.*?</email>}
	{<option>.*?</option>}
	{<hostid>.*?</hostid>}
	{<replaceable>.*?</replaceable>}
	{<devicename>.*?</devicename>}
	{<keyword>.*?</keyword>}
	{<varname>.*?</varname>}
	{<envar>.*?</envar>}
	{<term>.*?</term>}
	{<makevar>.*?</makevar>}
	{<maketarget>.*?</maketarget>}
	{<quote>.*?</quote>}
	{<title>.*?</title>}
	{<link.*?>.*?</link>}
	{<ulink.*?>.*?</ulink>}
	{<errorname>.*?</errorname>}
	{<groupname>.*?</groupname>}
	{<entry>.*?</entry>}
	{<emphasis>.*?</emphasis}
	{<citerefentry>.*?</citerefentry>}
	{<sect1info>.*?</sect1info>}
	{<sect2info>.*?</sect2info>}
	{<citetitle>.*?</citetitle>}
	{<.*?>}
    }
    
    foreach exp $reg {
	regsub -all $exp $text "" text
    }
    return $text
}

proc main { argc argv0 argv } {
    global dictionary ifd

    set s 0
    set w 0

    set tcount 0
    set terrs  0

    set stats(tread)  0
    set stats(tspell) 0
    set stats(tparse) 0

    set ifd [open "| ispell -a -H" a+]
    gets $ifd

    set argv [split $argv " "]

    if {$argc < 1} {
	puts stderr "Usage: $argv0 \[-s | -ss\] \[-w WIDTH\] <file1> \[file2\] \[file3\]"
	puts stderr ""
	puts stderr "  -w WIDTH   Screen Width"
	puts stderr "  -s stats"
	puts stderr "  -ss super stats"
	return
    }

    foreach item $argv {
	switch -- $item {
	    "-s" {
		set s 1
	    }
	    "-ss" {
		set s 2
	    }
	    "-w" {
		set w 1
	    }
	    default {
		if {$w} {
		    if {[regexp -all {[0-9]} $item] < [string length $item]} {
			puts stderr "$item must be a number"
		    } elseif { $width < 20 } {
			puts stderr "$item too small width"
		    } else {
			set width $item
		    }
		    set w 0
		} elseif {[file isfile $item] && [file readable $item]} {
		    lappend files $item
		} else {
		    puts stderr "$item does not exist/permission denied."
		}
	    }
	}
    }

    if {![info exists files]} {
	puts stderr "No suitable files found."
	return
    }

    foreach file $files {
	puts stderr "Checking $file"

	set stats(read) [secs]
	set ffd [open $file r]
	set text [read $ffd]
	close $ffd
	set stats(read) [expr [secs] - $stats(read)]
	incr stats(tread) $stats(read)

	set stats(parse) [secs]
	set text [numbertext $text]
	set text [removetags $text]
	set text [split $text \n]
	set stats(parse) [expr [secs] - $stats(parse)]
	incr stats(tparse) $stats(parse)

	set count 0
	set errs  0

	set stats(spell) [secs]

	foreach line $text {
	    set sline [lrange [split $line " "] 1 end]

	    foreach word $sline {
		set word [string trim $word]; set cword ""
		if {$word != ""} {
		    if {[regexp {^([A-Za-z]*?)(\.|,)$} $word null nword]} {
			set cword $nword
		    } elseif {[regexp {^[A-Za-z]*$} $word]} {
			set cword $word
		    }
		    if {$cword != "" && [lsearch $dictionary [string tolower \
			$cword]] == -1 && ![speller $cword]} {
			if {[info exists width]} {
			    set nline [string range $line 0 [expr $width - 17]]
			} else {
			    set nline $line
			}
			puts "[format "%-15s" $cword] $nline"
			incr errs
			incr terrs
		    }
		}
		incr count
		incr tcount
	    }
	}
	set stats(spell) [expr [secs] - $stats(spell)]
	incr stats(tspell) $stats(spell)

	switch -- $s {
	    "1" {
		puts stderr "Checked $file -- $count words, $errs errors"
	    }
	    "2" {
		puts stderr "Checked $file -- $count words, $errs errors"
		puts stderr "  read: $stats(read)s"
		puts stderr "  parse: $stats(parse)s"
		puts stderr "  spell: $stats(spell)s"
	    }
	}
    }
    if {$s == 2 && [llength $files] > 1} {
	puts stderr "Totals -- $tcount words, $terrs errors"
	puts stderr "  read: $stats(tread)s"
	puts stderr "  parse: $stats(tparse)s"
	puts stderr "  spell: $stats(tspell)s"
    }
}

main $argc $argv0 $argv